""" This module contains a Conn class to connect and
retrieve data from Eulerian Technologies API
"""

import hashlib
import inspect
import time

from ..internal import _request


class Conn:
    """Setup the connexion to Eulerian Technologies API.

    Parameters
    ----------
    gridpool_name: str, obligatory
        Your assigned grid in Eulerian Technologies platform

    datacenter: str, obligatory
        Your assigned datacenter (com for Europe, ca for Canada) in Eulerian Technologies platform

    api_key: str, obligatory
        Your Eulerian Technologies user account API key

    print_log: str, optional
        Set to False to hide log message
        Default: True

    Returns
    -------
        Class is instantiated
    """
    _cached_creds = set()

    def __init__(
            self,
            gridpool_name: str,
            datacenter: str,
            api_key: str,
            print_log: bool = True
    ):
        if not isinstance(print_log, bool):
            raise TypeError("print_log should be a boolean type")

        if not isinstance(gridpool_name, str) or len(gridpool_name) == 0:
            raise TypeError("gridpool_name should be a non-null string type")

        if not isinstance(datacenter, str) or len(datacenter) == 0:
            raise TypeError("datacenter should be a non-null string type")

        if not isinstance(api_key, str) or len(api_key) == 0:
            raise TypeError("api_key should be a non-null string type")

        self._cached_credentials = set()
        self._datacenter = datacenter
        self._gridpool_name = gridpool_name
        self._base_url = f"https://{gridpool_name}.api.eulerian.{datacenter}"
        self._api_v2 = f"{self._base_url}/ea/v2"
        self._api_key = api_key
        self._http_headers = {"Authorization": f"Bearer {api_key}"}
        self._print_log = print_log

        if self._creds_hex_digest in Conn._cached_creds:
            self._log("Credentials found in cache")

        else:
            self._check_credentials()
            Conn._cached_creds.add(self._creds_hex_digest)
            self._log("New credentials ok")

    # Import class methods
    from ._download_datamining import download_datamining
    from ._download_edw import download_edw
    from ._download_realtime_report import download_realtime_report
    from ._download_flat_realtime_report import download_flat_realtime_report, _get_all_paths, _all_paths_to_df
    from ._download_flat_overview_realtime_report import download_flat_overview_realtime_report

    @property
    def _creds_hex_digest(self) -> str:
        concate_creds_values = ''.join([
            self._datacenter,
            self._gridpool_name,
            self._api_key
        ])
        hash_object = hashlib.md5(concate_creds_values.encode())
        creds_hex_digest = hash_object.hexdigest()
        return creds_hex_digest

    def _check_credentials(self) -> None:
        """Check credentials validity"""
        # raise an error if API error or fail to load as json
        overview_url = f"{self._api_v2}/er/account/authtree.json"
        _request._to_json(
            request_type="get",
            url=overview_url,
            headers=self._http_headers,
            print_log=self._print_log
        )

    def _log(
            self,
            log: str,
    ) -> None:
        """ A simple logging mechanism
        Parameters
        ----------
        log: str, obligatory
            Log message to be displayed

        Returns
        -------
            None, print a log if print_log is True
        """
        if not isinstance(log, str):
            raise TypeError("log should be str dtype")

        if self._print_log:
            stack = inspect.stack()
            frame = stack[1]
            caller_func = frame.function
            caller_mod = inspect.getmodule(frame[0])
            log_msg = f"{time.ctime()}:{caller_mod.__name__}:{caller_func}: {log}"
            print(log_msg)

        return None

    def get_view_id_name_map(
            self,
            website_name: str
    ) -> dict:
        """ Fetch attribution rules

        Parameters
        ----------
        website_name: str, obligatory
            Your targeted website_name in Eulerian Technologies platform

        Returns
        -------
        dict
            A dict as { "view_id" : "view_name", ...}
        """
        if not isinstance(website_name, str):
            raise TypeError("website_name should be a string")

        view_url = f"{self._api_v2}/ea/{website_name}/db/view/get_all_name.json"
        view_json = _request._to_json(
            request_type="get",
            url=view_url,
            headers=self._http_headers,
            print_log=self._print_log
        )

        view_id_idx = view_json["data"]["fields"].index({"name": "view_id"})
        view_name_idx = view_json["data"]["fields"].index({"name": "view_name"})
        views = {view[view_id_idx]: view[view_name_idx] for view in view_json["data"]["rows"]}

        if "0" not in views:
            views["0"] = "last channel"

        return views

    def get_website_by_name(
            self,
            website_name: str
    ) -> dict:
        """ Fetch website properties

        Parameters
        ----------
        website_name: str, obligatory
            Your targeted website_name in Eulerian Technologies platform

        Returns
        -------
        dict
            A dict as { "website_prop" : "website_prop_value" }
        """
        if not isinstance(website_name, str):
            raise TypeError("website_name should be a string")

        website_url = f"{self._api_v2}/ea/{website_name}/db/website/get_me.json"
        website_json = _request._to_json(
            request_type="get",
            url=website_url,
            params={"output-as-kv": 1},
            headers=self._http_headers,
            print_log=self._print_log
        )

        d_website = website_json["data"]["rows"][0]
        if not isinstance(d_website, dict):
            raise TypeError(f"d_website={d_website} should be a dict dtype")
        return d_website

    def get_mdevicetype_id_name_map(
            self,
            website_name: str,
    ) -> dict:
        """ Fetch id and name properties of mdevicetype class

        Parameters
        ----------
        website_name: str, obligatory
            Your targeted website_name in Eulerian Technologies platform

        Returns
        -------
        dict
            A dict as { "mdevicetype_id" : "mdevicetype_name" }
        """
        if not isinstance(website_name, str):
            raise TypeError("website_name should be a string")

        mdevicetype_url = f"{self._api_v2}/ea/{website_name}/db/mdevicetype/getall.json"
        _json = _request._to_json(
            request_type="get",
            url=mdevicetype_url,
            params={"output-as-kv": 1},
            headers=self._http_headers,
            print_log=self._print_log
        )

        return {
            _json["data"]["rows"][i]["mdevicetype_id"]: _json["data"]["rows"][i]["mdevicetype_name"]
            for i in range(len(_json["data"]["rows"]))
        }

    def get_ordertype_id_name_map(
            self,
            website_name: str,
    ) -> dict:
        """ Fetch id and name properties of the ordertype class

        Parameters
        ----------
        website_name: str, obligatory
            Your targeted website_name in Eulerian Technologies platform

        Returns
        -------
        dict
            A dict as { "ordertype_id" : "ordertype_name" }
        """
        if not isinstance(website_name, str):
            raise TypeError("website_name should be a string")

        ordertype_url = f"{self._api_v2}/ea/{website_name}/db/ordertype/searchvisible.json"
        _json = _request._to_json(
            request_type="get",
            url=ordertype_url,
            params={"limit": 500, "output-as-kv": 1},
            headers=self._http_headers,
            print_log=self._print_log
        )

        return {
            _json["data"]["rows"][i]["ordertype_id"]: _json["data"]["rows"][i]["ordertype_key"]
            for i in range(len(_json["data"]["rows"]))
        }

    def get_ordertypecustom_id_name_map(
            self,
            website_name: str,
    ) -> dict:
        """ Fetch id and name properties of the ordertypecustom class

        Parameters
        ----------
        website_name: str, obligatory
            Your targeted website_name in Eulerian Technologies platform

        Returns
        -------
        dict
            A dict as { "ordertypecustom_id" : "ordertypecustom_name" }
        """
        if not isinstance(website_name, str):
            raise TypeError("website_name should be a string")

        ordertypecustom_url = f"{self._api_v2}/ea/{website_name}/db/ordertypecustom/searchvisible.json"
        _json = _request._to_json(
            request_type="get",
            url=ordertypecustom_url,
            params={"limit": 100, "output-as-kv": 1},
            headers=self._http_headers,
            print_log=self._print_log
        )

        return {
            _json["data"]["rows"][i]["ordertypecustom_id"]: _json["data"]["rows"][i]["ordertypecustom_name"]
            for i in range(len(_json["data"]["rows"]))
        }

    def get_profile_id_name_map(
            self,
            website_name: str,
    ) -> dict:
        """ Fetch id and name properties of the profile class

        Parameters
        ----------
        website_name: str, obligatory
            Your targeted website_name in Eulerian Technologies platform

        Returns
        -------
        dict
            A dict as { "profile_id" : "profile_name" }
        """
        if not isinstance(website_name, str):
            raise TypeError("website_name should be a string")

        profile_url = f"{self._api_v2}/ea/{website_name}/db/profile/search.json"
        _json = _request._to_json(
            request_type="get",
            url=profile_url,
            params={"limit": 100, "output-as-kv": 1},
            headers=self._http_headers,
            print_log=self._print_log
        )

        return {
            _json["data"]["rows"][i]["profile_id"]: _json["data"]["rows"][i]["profile_name"]
            for i in range(len(_json["data"]["rows"]))
        }



"""Internal os helper"""

import os
from ._log import _log


def _remove_file(
        path2file: str,
) -> None:
    """ Remove the given path2file

    Parameters
    ----------
    path2file : str, obligatory
        The output full path to file

    Returns
    -------
        A boolean
    """
    if not isinstance(path2file, str):
        raise TypeError(f"path2file={path2file} should be a str type")

    if os.path.isfile(path2file):
        os.remove(path2file)

    return None


def _create_directory(
        output_directory: str
) -> None:
    """ Create directory if it does not exists on the file system

    Parameters
    ----------
    output_directory : str, obligatory
        The output directory to be created on the file system

    """
    if output_directory:
        if not isinstance(output_directory, str):
            raise TypeError("output_directory should be a str type")

        if not os.path.exists(output_directory):
            os.mkdir(output_directory)

    return None



"""This module allows to download realtime report data
from the Eulerian Technologies API"""

import pandas as pd

from eanalytics_api_py.internal import _request


def download_realtime_report(
        self,
        website_name: str,
        report_name: list,
        payload: dict,
):
    """ Fetch realtime report data into a pandas dataframe

    Parameters
    ----------
    website_name : str, mandatory
        Your targeted website_name in Eulerian Technologies platform

    report_name: str, mandatory

    payload : dict, mandatory
        The realtime report payload

    Returns
    -------
    pd.DataFrame()
        A pandas dataframe
    """

    if not isinstance(website_name, str):
        raise TypeError("website_name should be a str type")

    if not isinstance(report_name, str):
        raise TypeError("report_name should be a str type")

    if not isinstance(payload, dict):
        raise TypeError("payload should be a dict type")

    if not payload:
        raise ValueError("payload should not be empty")

    report_url = f"{self._api_v2}/ea/{website_name}/report/realtime/{report_name}.json"
    payload['ea-switch-datetorow'] = 1  # include the date in each row
    payload['ea-enable-datefmt'] = "%s"  # format the date as an epoch timestamp

    report_json = _request._to_json(
        request_type="get",
        url=report_url,
        params=payload,
        headers=self._http_headers,
        print_log=self._print_log
    )

    fields = [field['name'] for field in report_json['data']['fields']]
    rows = report_json['data']['rows']

    df = pd.DataFrame(
        columns=fields,
        data=rows,
    )

    df["website"] = website_name  # useful when querying multiple websites will be implemented
    df["date"] = pd.to_datetime(df["date"], unit='s')

    return df



""" Generic load from csv file into pandas DataFrame with transformation """

import re as _re
import pandas as _pd

def csv_files_2_df(
    path2files : list,
    sep=';',
    quotechar='"',
    compression='gzip',
    encoding='utf-8',
    **kwargs
):
    """ Load a list of csv files into a pandas dataframe

    Parameters
    ----------
    path2files : list, obligatory
        The targeted list of path2files

    sep : str, obligatory
        The csv sep char
        Default: ';'

    quotechar : str, optional
        The csv quote char
        Default: '"'

    compression : str, optional
        The file compression algorithm
        Default: 'gzip'

    encoding: str, optional
        The file encoding
        Default: 'utf-8'

    **kwargs:
        Keyword arguments for pd.read_csv function


    Returns
    -------
    pd.Dataframe
        Pandas dataframe object
    """
    l_df = []

    if isinstance(path2files, str):
        path2files = [ path2files ]

    if not isinstance(path2files, list):
        raise TypeError("path2files should be either a str or a list type")

    for path2file in path2files:
        df = _pd.read_csv(
            path2file,
            sep=sep,
            quotechar=quotechar,
            compression=compression,
            encoding=encoding,
            index_col=None,
            header=0,
            **kwargs,
        )
        l_df.append(df)

    df_concat = _pd.concat(l_df, axis=0, ignore_index=True)
    __set_df_col_dtypes(df_concat)

    # if view-id=0 similuate attrib view col
    if "viewchannel_lvl_p0" not in df_concat.columns:
        for col_name in df_concat:
            if col_name.startswith("channel_lvl0_"):
                newcol_name = _re.sub(
                    pattern=r"^channel_lvl0_(.+)$",
                    repl=r"viewchannel_\g<1>",
                    string=col_name
                )
                df_concat[newcol_name] = df_concat[col_name]

    # rename to match new convention
    else:
        d_col_rename_map = {}
        for col_name in df_concat:
            new_colname = _re.sub(
                pattern=r"^(viewchannel)_lvl_(.+)$",
                repl=r"\g<1>_\g<2>",
                string=col_name
            )
            d_col_rename_map[col_name] = new_colname
        df_concat = df_concat.rename(
            columns=d_col_rename_map
        )

    return df_concat

def __set_df_col_dtypes( df : _pd.DataFrame() ):
    """ Load a list of csv files into a pandas dataframe

    Parameters
    ----------
    df : pd.DataFrame, obligatory

    Returns
    -------
    pd.Dataframe
        Pandas dataframe object with dtype updated
    """
    if not isinstance(df, _pd.DataFrame):
        raise TypeError("argument should be a pd.DataFrame")

    # category
    for col_name in [
        "order_status",
        "channel_lvl_via",
        "channel_lvl_profile",
        "ordertype_key",
        "orderproduct_ref",
        "orderproduct_name",
        "productgroup_name",
    ]:
        if col_name in df.columns:
            df[col_name] = df[col_name].astype('category')

    # int-16
    for col_name in [
        "a_channel_sz",
        "a_orderproduct_sz",
        "channel_lvl_position",
        "product_position",
        "orderproduct_quantity",
    ]:
        if col_name in df.columns:
            df[col_name] = df[col_name].astype('int16')

    for col_name in df.columns:
        if col_name.startswith('channel_lvl_p'):
            df[col_name] = df[col_name].astype('category')
        elif df[col_name].dtype == 'object':
            df[col_name] = df[col_name].astype('category')
        elif df[col_name].dtype == 'int64':
            df[col_name] = df[col_name].astype('int32')
        elif df[col_name].dtype == 'float64':
            df[col_name] = df[col_name].astype('float32')

    return df



"""Wrapper around ipywidgets for easier manipulation """

import ipywidgets as _widgets
# https://ipywidgets.readthedocs.io/en/latest/examples/Widget%20List.html

def dropdown(
    d : dict,
    label = 'key',
    **kwargs,
):
    """ Display a dropdown for a given dictionnary

        Parameters
        ----------
        d : dict, mandatory,

        **kwargs: arguments for widgets.Dropdown

        label: str, optional, allowed='key', 'value'
            display key as label or value as label
            default: 'key'

    Returns
    -------
    HTML dropdown to toggle cell visibility.
    """
    if not isinstance(d, dict):
        raise TypeError("Argument should be a dictionnary")

    if label == 'key':
        list_of_tuples = [(k, v) for k, v in d.items()]
    elif label == 'value':
        list_of_tuples = [(v, k) for k, v in d.items()]
    else:
        raise ValueError("Allowed labels=key, value")

    return _widgets.Dropdown(
        options=list_of_tuples,
        **kwargs
    )



"""Internal logging module"""

import inspect
import time


def _log(
        log: str,
        print_log: bool = True
) -> None:
    """ A simple logging mechanism
    Parameters
    ----------
    log: str, obligatory
        Log message to be displayed

    print_log: bool, optional
        Whether to print the log message or not
    Returns
    -------
        None, print a log if print_log is True
    """
    if not isinstance(log, str):
        raise TypeError("log should be str dtype")

    if not isinstance(print_log, bool):
        raise TypeError("print_log should be a bool dtype")

    if print_log:
        stack = inspect.stack()
        frame = stack[1]
        caller_func = frame.function
        caller_mod = inspect.getmodule(frame[0])
        log_msg = f"{time.ctime()}:{caller_mod.__name__}:{caller_func}: {log}"
        print(log_msg)

    return None



"""This module allows to download flat realtime report data
from the Eulerian Technologies API"""

from pprint import pprint

import pandas as pd

from eanalytics_api_py.internal import _request


def download_flat_realtime_report(
        self,
        date_from: str,
        date_to: str,
        website_name: str,
        report_name: list,
        path_dim_map: list,
        kpi: list,
        date_scale: str = '',
        view_id: int = 0,
        mdevicetype_id: list = None,
        ordertype_id: list = None,
        ordertypecustom_id: list = None,
        profile_id: list = None,
) -> pd.DataFrame:
    """ Fetch realtime report data into a pandas dataframe

    Parameters
    ----------
    date_from: str, mandatory
        mm/dd/yyyy

    date_to: str, mandatory
        mm/dd/yyyy

    website_name: str, mandatory
        Your targeted website_name in Eulerian Technologies platform

    report_name: str, mandatory

    path_dim_map: list, mandatory
        List of paths to drill down

    kpi: list, mandatory
        List of kpis to request

    date_scale: str, optional
        Split data for a given scale
        Allowed values H, D, W, M

    view_id: int, optional
        Between 0 and 9

    mdevicetype_id: list, optional
        List of mdevicetype_id to filter on

    ordertype_id: list, optional
        List of ordertype_id to filter on

    ordertypecustom_id: list, optional
        List of ordertypecustom_id to filter on

    profile_id: list, optional
        List of profile_id to filter on

    Returns
    -------
    pd.DataFrame()
        A pandas dataframe
    """

    if not isinstance(date_from, str):
        raise TypeError("date_from should be a string dtype")

    if not isinstance(date_to, str):
        raise TypeError("date_to should be a string dtype")

    if not isinstance(website_name, str):
        raise TypeError("website_name should be a string dtype")

    if not isinstance(report_name, str):
        raise TypeError("report_name should be a string dtype")

    if not isinstance(path_dim_map, dict):
        raise TypeError("path_dim_map should be a dict dtype")

    if not isinstance(kpi, list):
        raise TypeError("kpi should be a list dtype")

    if not isinstance(date_scale, str):
        raise TypeError("date_scale be a str dtype")

    payload = {
        'date-from': date_from,
        'date-to': date_to,
        'ea-switch-datetorow': 1,  # include the date in each row
        'ea-enable-datefmt': "%s",  # format the date as an epoch timestamp
        'ea-columns': "id," + ",".join(kpi),
    }
    if mdevicetype_id:
        if not isinstance(mdevicetype_id, list):
            raise TypeError("mdevicetype_id should be a list")

        mdevicetype_map = self.get_mdevicetype_id_name_map(website_name)
        for _id in mdevicetype_id:
            if _id not in mdevicetype_map:
                raise ValueError(f"mdevicetype_id={_id} not allowed. Allowed={', '.join(mdevicetype_map.keys())}")
        payload["ea-subk"] = ",".join(mdevicetype_id)

    if ordertype_id:
        if not isinstance(ordertype_id, list):
            raise TypeError("ordertype_id should be a list")
        ordertype_map = self.get_ordertype_id_name_map(website_name)
        for _id in ordertype_id:
            if _id not in ordertype_map:
                raise ValueError(f"ordertype_id={_id} not allowed. Allowed={', '.join(ordertype_map.keys())}")
        payload["shoppingcart-k1"] = ",".join(ordertype_id)

    if ordertypecustom_id:
        if not isinstance(ordertypecustom_id, list):
            raise TypeError("ordertypecustom_id should be a list")
        ordertypecustom_map = self.get_ordertypecustom_id_name_map(website_name)
        for _id in ordertypecustom_id:
            if _id not in ordertypecustom_map:
                raise ValueError(
                    f"ordertypecustom_id={_id} not allowed. Allowed={', '.join(ordertypecustom_map.keys())}")
        payload["shoppingcart-k1-custom"] = ",".join(ordertypecustom_id)

    if profile_id:
        if not isinstance(profile_id, list):
            raise TypeError("profile_id should be a list")
        profile_map = self.get_profile_id_name_map(website_name)
        for _id in profile_id:
            if _id not in profile_map:
                raise ValueError(f"profile_id={_id} not allowed. Allowed={', '.join(profile_map.keys())}")
        payload["visitpprofil-k1"] = ",".join(profile_id)

    l_allowed_scale = ["H", "D", "W", "M"]
    if date_scale and date_scale not in l_allowed_scale:
        raise ValueError(f"date_scale={date_scale} not allowed. Allowed: {', '.join(l_allowed_scale)}")

    view_id = str(view_id)
    view_map = self.get_view_id_name_map(website_name)
    if view_id not in view_map:
        raise ValueError(f"view_id={view_id} not found. Allowed: {', '.join(view_map.keys())}")

    payload["view-id"] = view_id

    d_website = self.get_website_by_name(website_name)
    url = f"{self._api_v2}/ea/{website_name}/report/realtime/{report_name}.json"

    l_df = []
    for path, l_dim in path_dim_map.items():
        if not isinstance(path, str):
            raise ValueError("path in path_dim_map should ba a str dtype")

        if not isinstance(l_dim, list):
            raise ValueError("dim in path_dim_map should ba a list dtype")

        l_path = path.split(".")
        l_path[0] = l_path[0] % int(d_website["website_id"])
        l_all_paths = self._get_all_paths(
            i=1,
            l_path=l_path,
            l_prev_path=[l_path[0]],
            url=url,
            payload=payload)

        sub_df = self._all_paths_to_df(
            url=url,
            l_path=l_all_paths,
            l_dim=l_dim,
            l_kpi=kpi,
            payload=payload,
            date_scale=date_scale)

        l_df.append(sub_df)

    df = pd.concat(
        objs=l_df,
        axis=0,
        ignore_index=True)

    return df


def _get_all_paths(
        self,
        i: int,
        l_path: list,
        l_prev_path: list,
        url: str,
        payload: dict,
):
    l_next_path = []
    for prev_path in l_prev_path:
        if not l_path[i].endswith("[%d]"):
            l_next_path.append(".".join([prev_path, l_path[i]]))

        else:
            payload["path"] = ".".join([prev_path, l_path[i].replace("[%d]", "")])
            _json = _request._to_json(
                url=url,
                request_type="get",
                headers=self._http_headers,
                params=payload,
                print_log=self._print_log
            )
            for _id in _get_ids(_json):
                l_next_path.append(".".join([prev_path, l_path[i] % int(_id)]))

    if i == len(l_path) - 1:
        return l_next_path
    i += 1
    return self._get_all_paths(
        i=i,
        l_path=l_path,
        l_prev_path=l_next_path,
        url=url,
        payload=payload)


def _all_paths_to_df(
        self,
        url: str,
        date_scale: str,
        l_path: [],
        l_dim,
        l_kpi,
        payload: {}
):
    payload["ea-columns"] = "name," + ",".join([*l_dim, *l_kpi])
    if date_scale:
        del(payload["ea-columns"])
        payload["date-scale"] = date_scale
        payload["dd-dt"] = ",".join([*l_dim, *l_kpi])

    l_slice_path = []
    l_df = []
    for i in range(len(l_path)):
        l_slice_path.append(l_path[i])
        if len(l_path) == 1 or (i and (i % 10 == 0 or i == len(l_path) - 1)):
            payload['path'] = ",".join(l_slice_path)
            _json = _request._to_json(
                url=url,
                request_type="get",
                params=payload,
                headers=self._http_headers,
                print_log=self._print_log
            )
            sub_df = pd.DataFrame(
                data=_json["data"]["rows"],
                columns=[d_field["name"] for d_field in _json["data"]["fields"]])

            l_df.append(sub_df)
            l_slice_path = []

    df_concat = pd.concat(
        objs=l_df,
        axis=0,
        ignore_index=True)

    return df_concat


def _get_ids(_json):
    for i, d_header in enumerate(_json["data"]["fields"]):
        if d_header["name"] == "id":
            return (row[i] for row in _json["data"]["rows"])



"""Http params grouped into dicts of category"""

product = {
    "with-orderproduct" : 1, #  purchased products
    "with-productparam" : 1, # purchased products parameters
    "with-productgroup" : 1, # purchased products groups
}

channel = {
    "with-channel-count" : 1, # number of marketing interactions
    "with-channel-level" : 1, # with previous marketing interactions
    "max-channel-info" : 14,  # marketing interactions detail level
    "max-channel-level" : 40, # marketing interaction history length
    "with-channel-via" : 1,   # attibution type pc/pw/r...
    "with-channel-date" : 1,  # marketing interaction date
    "with-channel-profile" : 1, # marketing interaction profile
    "with-channel-mdevicetype" : 1, # marketing interaction device type
    "with-first-channel" : 1, # with first marketing interaction
    "with-last-channel" : 1, # with last marketing interaction
    "with-firstlast-channel-lag" : 1, # delay (seconds) between first and last interaction
}

tpchannel = {
    "with-tpchannel-count" : 1,
    "with-tpchannel-level" : 1,
    "max-tpchannel-info" : 9,
    "max-tpchannel-level" : 40,
    "with-tpchannel-via" : 1,
    "with-tpchannel-date" : 1,
    "with-tpchannel-profile" : 1,
    "with-tpchannel-mdevicetype" : 1,
    "with-first-tpchannel" : 1,
    "with-last-tpchannel" : 1,
    "with-firstlast-tpchannel-lag" : 1,
}

order = {
    "with-ordertype" : 1, # order type
    "with-ordertypecustom" : 1, # order custom type
    "with-orderpayment" : 1, # order payment
    "with-orderpage" : 1, # path seen during the order session
    "with-orderpageurl" : 1, # url seen during the order session
    "with-orderevent" : 1, #  event triggered during the order session
    "with-scartdate" : 1, # shopping cart date
    "with-scartvalidmargin" : 1, # order margin
    "with-realscartvalidmargin" : 1, # real order margin
    "with-scartvalidamount" : 1, # order amount
    "with-realscartvalidamount" : 1, # real order amount
}

analytics = {
    "with-extparam" : 1,
    "with-geoloc" : 1, # country Coded on 2 ISO-3166-1 characters
    "with-browser" : 1, # browser name
    "with-lang" : 1, # browser language
    "with-os" : 1, # operating system
    "with-isp" : 1, # internet service provider
    "with-action" : 1, # action executed before purchasing
    "with-mdevicetype" : 1, # type of device
    "with-blockpromo" : 1, # last clicked auto promotion
    "with-visitdate" : 1, # visit date
    "with-visitlag" : 1, # duration of the visit
    "with-cgiparam" : 1, # custome parameters
    "with-rebuyer-prevnb" : 1, # previous number of orders
    "with-rebuyer-prevamount" : 1, # previous revenue amount
    "with-rebuyer-prevtm" : 1, # date of previous order
    "with-rebuyer-prevtm-epoch" : 1, # date EPOCH of previous order
}

user = {
    "with-email" : 1,
    "with-ip" : 1,
    "with-buyerid" : 1,
    "with-userid" : 1,
    "with-last-rp-ea-android-adid" : 1,
    "with-last-rp-ea-ios-idfa" : 1,
}

user_info = {
    "with-cluster" : 1, # Segment names
    "with-iduserparam" :1, # CRM parameters
}



"""Interact with Jupyter Notebook HTML code"""

from IPython.display import HTML as _HTML

def button_hide_cells(
    tagname_divclass : dict,
    force_hide = False
):
    """ Hide cells matching the given tags

    Button to hide specific cells, while displaying config cells to the end users.

    Parameters
    ----------
    tag_name_divclass : a dict as { "tagname" : "divclass" }, mandatory
        "tag_name": name of the tag to hide
        "divclass : "input", "output", "input_output"


    force_hide: bool, optional
        Set to True to hide matching cell when executing the function
        Default :False

    Returns
    -------
    HTML button to toggle cell visibility. To use with the "display" command
    """
    if not isinstance(tagname_divclass, dict):
        raise TypeError("tagname_divclass should be a dict")

    force_hide = 1 if force_hide else 0 # javascript compatibility
    allowed_divclass = ["input", "output", "input_output"]

    for divclass in tagname_divclass.values():
        if divclass not in allowed_divclass:
            raise ValueError(f"divclass={divclass} not allowed, use: {', '.join(allowed_divclass)}")

    html = _HTML(f'<script>\
        hide_code = true;\
        function code_toggle(fromForm=false) {{\
            if ( ( !fromForm || typeof(fromForm) === "function" ) && ( !{force_hide} ) ) {{\
                return 0;\
            }}\
            $( "div.input" ).each(function() {{\
                var divInput = this;\
                $( this ).find("span.cell-tag").each(function() {{\
                    var tagname = $( this ).text();\
                    if ( tagname in {tagname_divclass}  ) {{\
                        var divname = {tagname_divclass}[tagname]; \
                        if ( hide_code ) {{\
                            if ( divname === "input" || divname === "input_output" ) $( divInput ).hide();\
                            if ( divname === "output" || divname === "input_output" ) $( divInput ).next().hide();\
                        }}\
                        else {{\
                            if ( divname === "input" || divname === "input_output" ) $( divInput ).show() ;\
                            if ( divname === "output" || divname === "input_output" )$( divInput ).next().show();\
                        }}\
                    }}\
                }});\
            }});\
            hide_code = !hide_code\
        }}\
        \
        $( document ).ready(code_toggle);\
        </script>\
        <form action="javascript:code_toggle(fromForm=true)">\
            <input type="submit" value="Code tag visibility">\
        </form>'
    )

    return html

def button_hide_input_cells():
    """ Toggle visibility of all input cells (div.input)

    Parameters
    ----------
    None

    Returns
    -------
    HTML button to toggle cell visibility. To use with the "display" command
    """
    html = _HTML(
        '''
        <script>
            code_show=false;
            function code_toggle() {
                 if (code_show){
                    $('div.input').hide();
                 }
                 else {
                    $('div.input').show();
                 }
                 code_show = !code_show
            }
            $( document ).ready(code_toggle);
        </script>
        <form action="javascript:code_toggle()">
            <input type="submit" value="Toggle input cells visiblity">
        </form>
        '''
    )
    return html

def remove_class_output_scroll():
    """ Remove all output cells scrollbar by removing the output_scroll class

    Parameters
    ----------
    None

    Returns
    -------
    HTML code to use inside the "display" command
    """
    js = "<script>$('.output_scroll').removeClass('output_scroll')</script>"
    return _HTML(js)



""" List of path for insumnary realtime report
Drill down by channel """

dim_px_map = {
    "media_key": "p0",
    "ope_name": "p1",
    "opealias_alias": "p1",
    "publisher_name": "p2",
    "publisheralias_alias": "p2",
    "mediaplan_name": "p6",
    "mediaplangroup_name": "p7",
    "submedia_name": "p8",
    "publishergroup_name": "p18",
    "publishertype_name": "p19",
    "submediagroup_name": "p20",

}

override_dim_map = {
    "publisher_name": "publisheralias_alias",
    "ope_name": "opealias_alias",
}

d_path = {
    "ADVERTISING": {
        "path": ["mcMEDIAINCOMING[%d]", "mcMEDIAAD.mcOPE"],
        "dim": list(dim_px_map.keys())
    },
    "MAILING": {
        "path": ["mcMEDIAINCOMING[%d]", "mcMEDIAML.mcOPE"],
        "dim": list(dim_px_map.keys()),
    },
    "AFFILIATION": {
        "path": ["mcMEDIAINCOMING[%d]", "mcMEDIAAF.mcOPE"],
        "dim": list(dim_px_map.keys()),
    },
    "SOCIAL": {
        "path": ["mcMEDIAINCOMING[%d]", "mcMEDIASC.mcOPE"],
        "dim": list(dim_px_map.keys()),
    },
    "PAID INCLUSION": {
        "path": ["mcMEDIAINCOMING[%d]", "mcMEDIAPI.mcOPE"],
        "dim": list(dim_px_map.keys()),
    },
    "TRUSTED FEEDS": {
        "path": ["mcMEDIAINCOMING[%d]", "mcMEDIATF.mcOPE"],
        "dim": list(dim_px_map.keys()),
    },
    "PARTNER": {
        "path": ["mcMEDIAINCOMING[%d]", "mcMEDIAPT.mcOPE"],
        "dim": list(dim_px_map.keys()),
    },
    "SPONSORED LINKS": {
        "path": ["mcMEDIAINCOMING[%d]", "mcMEDIASL.mcOPE"],
        "dim": list(dim_px_map.keys()),
    },
}



""" Load datamining csv file into pandas DataFrame
Specific transformation available for easier analysis
"""

import re as _re

import pandas as _pd
import numpy as np
from .generic import csv_files_2_df

def deduplicate_touchpoints(
    source : list,
    sep=';',
    quotechar='"',
    compression='gzip',
    encoding='utf-8',
    **kwargs
):
    """ Deduplicate marketing touchpoints

        Parameters
        ----------
        source : list of path2file OR pandas DataFrame

        prdref_colname : str, optional

        sep : str, obligatory
            The csv sep char
            Default: ';'

        quotechar : str, optional
            The csv quote char
            Default: '"'

        compression : str, optional
            The file compression algorithm
            Default: 'gzip'

        encoding: str, optional
            The file encoding
            Default: 'utf-8'

        **kwargs:
            Keyword arguments for pd.read_csv function


    Returns
    -------
    pd.Dataframe
        Pandas DataFrame object
    """

    if isinstance(source, str):
        source = [ source ]


    if isinstance(source, list):
        df = csv_files_2_df(
            path2files=source,
            sep=sep,
            quotechar=quotechar,
            compression=compression,
            encoding=encoding,
            **kwargs,
        )

    elif isinstance(source, _pd.DataFrame):
        df = source

    else:
        raise ValueError("source should be a path2file or a pandas DataFrame")

    columns = []
    stubnames = []

    for col_name in df.columns:
        if col_name.startswith('channel_lvl'):
            col_name = _re.sub(
                pattern=r'^(channel)_lvl(\d+)_(.+)$',
                repl=r"\g<1>_\g<3>_\g<2>",
                string=col_name
            )
            prefix = _re.sub(
                pattern=r'^(channel_.+_)(\d+)$',
                repl=r"\g<1>",
                string=col_name
            )
            if prefix not in stubnames:
                stubnames.append(prefix)
        columns.append(col_name)

    df.columns = columns

    df = _pd.wide_to_long(
        df,
        stubnames=stubnames,
        i="order_ref",
        j="channel_position",
    ).reset_index()

    columns = []
    for col_name in df.columns:
        if col_name.startswith('channel_') and col_name.endswith('_'):
            col_name = col_name[:-1]
        columns.append(col_name)

    df.columns = columns
    df = df[~df.channel_p0.isin(['-',0,None, '0',np.nan])]

    return df

def deduplicate_products(
    source : list,
    sep=';',
    quotechar='"',
    compression='gzip',
    encoding='utf-8',
    **kwargs
):
    """ Deduplicate product params columns

    In eulerian datamining, each product related column is duplicated
    by the number of products.
    The goal is to perform the following transformation (example):
        product-color # 1,  product-color # 2, product-color # 3
    Into
        product-color

    This duplicates row for every product and normalize the product param columns.
    Rows having an empty product ref will be discarded.

    Parameters
    ----------
    source : list, obligatory
        The targeted list of path2files or pandas DataFrame

    prdref_colname : str, optional

    sep : str, obligatory
        The csv sep char
        Default: ';'

    quotechar : str, optional
        The csv quote char
        Default: '"'

    compression : str, optional
        The file compression algorithm
        Default: 'gzip'

    encoding: str, optional
        The file encoding
        Default: 'utf-8'

    **kwargs:
        Keyword arguments for pd.read_csv function


    Returns
    -------
    pd.Dataframe
        Pandas dataframe object
    """

    if isinstance(source, str):
        source = [ source ]

    if isinstance(source, list):
        df = csv_files_2_df(
            path2files=source,
            sep=sep,
            quotechar=quotechar,
            compression=compression,
            encoding=encoding,
            **kwargs,
        )

    elif isinstance(source, _pd.DataFrame):
        df = source

    else:
        raise ValueError("source should be a path2file or a pandas DataFrame")


    stubnames = []
    columns = []

    prd_colnames = [
        "orderproduct_name",
        "orderproduct_quantity",
        "orderproduct_ref",
        "orderproduct_sellprice",
        "productgroup_margin",
        "productgroup_name",
        "productparam_"
    ]

    columns = []

    # orderproduct_name_i
    for col_name in df.columns:
        for prd_col in prd_colnames:
            if col_name.startswith(prd_col):
                l_col_name = col_name.split("_")
                stubname = "_".join(l_col_name[:-1]) # orderproduct_name
                idx = l_col_name[-1] # i
                col_name = stubname+idx #orderproduct_namei
                if stubname not in stubnames:
                    stubnames.append(stubname)
        columns.append(col_name)
    df.columns = columns

    for stubname in stubnames:
        if stubname in columns:
            raise ValueError(f"stubname={stubname} should not equal to a column name")

    df = _pd.wide_to_long(
        df=df,
        stubnames=stubnames,
        i="order_ref",
        j="product_position"
    ).reset_index()
    df = df[~df.orderproduct_ref.isin(['-',0,None, '0',np.nan])]

    return df



from setuptools import setup, find_packages
from os import path

this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

setup(
    author='Florian',
    author_email='f.dauphin@eulerian.com',
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)",
        "Operating System :: OS Independent"
    ],
    description='Locally download a datamining dataset from the Eulerian Technologies API',
    download_url='https://github.com/EulerianTechnologies/eanalytics-api-py/archive/master.zip',
    install_requires=[
        'requests>=2.23.0',
        'ijson>=3.0.4',
        'pandas>=1.0.3',
        'ipython>=7.16.1',
        'ipywidgets>=7.5.1',
        'numpy>=1.19.1',
        'pytest>=6.1.1'
    ],
    keywords=[
        'eulerian',
        'datamining',
        'download',
        'dw',
        'datawharehouse',
        'eanalytics',
        'jupyter',
        'notebook',
        'data science'
    ],
    long_description=long_description,
    long_description_content_type='text/markdown',
    name='eanalytics_api_py',
    packages=find_packages(exclude=("tests",)),
    platforms=['any'],
    python_requires='>=3.6',
    url='https://github.com/EulerianTechnologies/eanalytics-api-py',
    version='0.1.31',
)



"""Request helper module"""

from pprint import pprint
import urllib
import os

import requests

from ._log import _log


def _to_json(
        request_type: str,
        url: str,
        headers: dict = None,
        params: dict = None,
        json_data: dict = None,
        print_log: bool = False
) -> dict:
    """ Make HTTP request and check for error

    Parameters
    ----------
    request_type: str, obligatory
        The type of request : get/post supported at the moment

    url: str, obligatory
        The url to request

    headers: dict, optional
        The dict to use as the request header

    params: dict, optional
        The dict to use as the request params (requests.get)

    json_data: dict, optional
        The dict to use as the request json params (requests.post)

    print_log: bool, optional
        Default: True
    Returns
    -------
        Request response loaded as JSON
    """
    if not isinstance(request_type, str):
        raise TypeError("request_type should be a string dtype")

    if not isinstance(url, str):
        raise TypeError("url should be a string dtype")

    if headers and not isinstance(headers, dict):
        raise TypeError("headers should be a dict dtype")

    if params and not isinstance(params, dict):
        raise TypeError("params should be a dict dtype")

    if json_data and not isinstance(json_data, dict):
        raise TypeError("json_data should be a dict dtype")

    request_map = {
        "get": requests.get,
        "post": requests.post
    }

    api_key = headers["Authorization"].split(" ")[1]
    log_url = url.replace("/ea/v2/", f"/ea/v2/{api_key}/")

    params = urllib.parse.urlencode(params, safe='/') if params else ''
    _log(
        log=f"url={log_url}?{params}",
        print_log=print_log
    )

    if request_type == "get":
        r = request_map["get"](
            url=url,
            headers=headers,
            params=params
        )

    elif request_type == "post":
        r = request_map["post"](
            url=url,
            headers=headers,
            json=json_data
        )

    else:
        allowed_requests_type = ["get", "post"]
        raise ValueError(f"request_type is not in {', '.join(allowed_requests_type)}")

    # if request cannot be converted into JSON
    try:
        r_json = r.json()

    # JSONDecodeError is a subclass of ValueError
    except ValueError as e:
        print(f"Could not convert'{r.text}' as json")
        raise e

    else:
        # potential errors from Eulerian Technologies API
        if "error" in r_json.keys() and r_json["error"] \
                or "status" in r_json.keys() and r_json['status'].lower() == "failed":
            print("JSON response from Eulerian Technologies API")
            pprint(r_json)
            raise SystemError(f"Error[{r.status_code}] from Eulerian Technologies API")

    return r_json


def _is_skippable(
        output_path2file: str,
        override_file: bool,
        print_log: bool = True
) -> bool:
    """ Load data from local file is exist and override_file is True

    Parameters
    ----------
    output_path2file : str, obligatory
        The output full path to file

    override_file : bool, obligatory
        Your assigned datacenter (com for Europe, ca for Canada)

    print_log: bool, optional
        Set to False to not display logs
        Default: True
    Returns
    -------
        True if we can fetch data directly from the file
    """
    if not isinstance(output_path2file, str):
        raise TypeError("output_path2file should be a string type")

    if not isinstance(override_file, bool):
        raise TypeError("override_file should be a string type")

    if os.path.isfile(output_path2file):
        if override_file:
            _log(
                log=f"Local file={output_path2file} will be overriden with new data",
                print_log=print_log)
            return False
        _log(
            log=f"Fetching data from local file={output_path2file}",
            print_log=print_log)
        return True

    _log(
        log=f"Local file={output_path2file} not found, downloading the data",
        print_log=print_log)
    return False



"""This module allows to download the raw data
from the Eulerian Data Warehouse"""

import re
import time
import urllib
import gzip
import csv

import requests
import ijson

from eanalytics_api_py.internal import _request, _log


def download_edw(
        self,
        query: str,
        status_waiting_seconds=5,
        ip: str = None,
        output_path2file=None,
        override_file=False,
        jobrun_id=None,
) -> str:
    """ Fetch edw data from the API into a gzip compressed file

    Parameters
    ----------
    query: str, obligatory
        EDW query

     status_waiting_seconds: int, optional
        Waiting time in seconds between each status query

    ip: str, optional
        Coma separated ip values
        Default: Automatically fetch your external ip address

    output_path2file: str, optional
        path2file where the data will be stored
        If not set, the file will be created in the current
            working directory with a default name

    override_file : bool, optional
        If set to True, will override output_path2file (if exists)
            with the new datamining content
        Default: False

    jobrun_id : str, optional
        The jobrun_id to download directly from a previously requested jobrun

    Returns
    -------
    str
        The output_path2file containing the downloaded datamining data
    """
    if not isinstance(query, str):
        raise TypeError("query should be a string")

    epoch_from_to_findall = re.findall(
        r'{\W+?(\d+)\W+?(\d+)\W+?}',  # { 1602958590 1602951590 }
        query
    )
    if not epoch_from_to_findall:
        raise ValueError(f"Could not read epoch_from and epoch_to from query=\n{query}")

    readers_findall = re.findall(r'\w+:\w+@[\w_-]+', query)  # ea:pageview@demo-fr
    if not readers_findall:
        raise ValueError(f"Could not read READER from query=\n{query}")

    # for valid filename
    readers = list(
        map(
            lambda s: s.replace(':', '_').replace('@', '_'),
            readers_findall
        )
    )

    if output_path2file:
        if not isinstance(output_path2file, str):
            raise TypeError("output_path2file should be a str type")

    else:
        output_path2file = '_'.join([
            "dw",
            self._gridpool_name,
            "_".join(epoch_from_to_findall[0]),
            "_".join(readers),
        ]) + ".cvs.gzip"

        if _request._is_skippable(
                output_path2file=output_path2file,
                override_file=override_file,
                print_log=self._print_log
        ):
            return output_path2file

    if ip:
        if not isinstance(ip, str):
            raise ValueError("ip should be a str type")

    else:
        self._log("No ip provided\
            \n Fetching external ip from https://api.ipify.org\
            \nif using a vpn, please provide the vpn ip\
        ")
        ip = requests.get(url="https://api.ipify.org").text

    edw_token_url = f"{self._api_v2}/er/account/get_dw_session_token.json"
    payload = {'ip': ip}

    edw_token_json = _request._to_json(
        request_type="get",
        url=edw_token_url,
        headers=self._http_headers,
        params=payload,
        print_log=self._print_log
    )

    edw_token = edw_token_json["data"]["rows"][0][0]
    edw_http_headers = {
        "Authorization": "Bearer " + edw_token,
        "Content-Type": "application/json"
    }

    # no jobrun_id provided, we send the query to get one
    if not jobrun_id:
        search_url = f"{self._base_url}:981/edw/jobs"

        edw_json_params = {
            "kind": "edw#request",
            "query": query
        }

        search_json = _request._to_json(
            request_type="post",
            url=search_url,
            json_data=edw_json_params,
            headers=edw_http_headers,
            print_log=self._print_log
        )
        jobrun_id = search_json['data'][0]

    status_url = f"{self._base_url}:981/edw/jobs/{jobrun_id}"

    if not isinstance(status_waiting_seconds, int) or status_waiting_seconds < 5:
        status_waiting_seconds = 5

    ready = False

    while not ready:
        self._log(f"Waiting for jobrun_id={jobrun_id} to complete")
        time.sleep(status_waiting_seconds)
        status_json = _request._to_json(
            request_type="get",
            url=status_url,
            headers=edw_http_headers,
            print_log=self._print_log
        )

        if status_json["status"] == "Done":
            ready = True
            download_url = status_json["data"][1]

    req = urllib.request.Request(
        url=download_url,
        headers=self._http_headers,
    )
    _stream_to_csv_gzip(
        req=req,
        path2file=output_path2file,
        search_url=search_url,
        edw_http_headers=edw_http_headers,
        print_log=self._print_log
    )

    self._log(f"Output csv path2file={output_path2file}")
    return output_path2file


def _stream_to_csv_gzip(
        req: urllib.request.Request,
        path2file: str,
        search_url: str,
        edw_http_headers: dict,
        print_log : bool = True
) -> None:
    """ Stream the Eulerian Data Warehouse data in csv gzipped file

    Parameters
    ----------
    req : urllib.request.Request, obligatory
        The request object to fetch from

    path2file : str, obligatory
        The path2file to put the data into

    search_url: str, obligatory
        The url to cancel the job

    edw_http_headers: dict, obligatory
        Edw authentification headers

    print_log: bool
        Set to False to hide logs
        Default=True
    """
    if req.has_header('Content-Length'):
        _log._log(
            log=f"Content-Length={int(req.get_header('Content-Length')) / (1024 * 1024):.2f} MBs",
            print_log=print_log)

    with gzip.open(
            filename=path2file,
            mode="wt"
    ) as csvfile:

        csvwriter = csv.writer(
            csvfile,
            delimiter=';'
        )

        columns = []
        with urllib.request.urlopen(req) as f:
            objects = ijson.items(f, "headers.schema.item")
            headers = (header for header in objects)
            for header in headers:
                columns.append(header[1])
            csvwriter.writerow(columns)

        with urllib.request.urlopen(req) as f:
            objects = ijson.items(f, "rows.item")
            rows = (row for row in objects)
            for row in rows:
                csvwriter.writerow(row)

    # Try to kill job, non blocking error
    with urllib.request.urlopen(req) as f:
        objects = ijson.items(f, "headers.uuid.item")
        uuids = (uuid for uuid in objects)
        for uuid in uuids:
            cancel_url = f"{search_url}/{uuid}/cancel"
            try:
                requests.post(
                    url=cancel_url,
                    headers=edw_http_headers
                )
            except Exception as e:
                _log._log(
                    log=f"Could not kill the process uuid={uuid}\n exception:{e}",
                    print_log=print_log)



""" This module allows to download a datamining
from the Eulerian Technologies API
"""

import urllib
import re
import os
import gzip
import time
from datetime import datetime, timedelta
import csv

import ijson

from eanalytics_api_py.internal import _os, _request


def download_datamining(
        self,
        website_name: str,
        datamining_type: str,
        payload=None,
        status_waiting_seconds=5,
        output_directory='',
        override_file=False,
        n_days_slice=31,
):

    """ Fetch datamining data from the API into a gzip compressed CSV file

    Parameters
    ----------
    website_name : str, obligatory
        Your targeted website_name in Eulerian Technologies platform

    datamining_type : str, obligatory
        The targeted datamining (isenginerequest, actionlogorder, scart ,estimate, order)

    payload : dict, optional
        The datamining payload that contains the requested data

    status_waiting_seconds: int, optional
        Waiting time in seconds between each status query

    output_directory : str, optional
        The local targeted  directory

    override_file : bool, optional
        If set to True, will override output_path2file (if exists)
            with the new datamining content
        Default: False

    n_days_slice: int, optional
        Split datamining query into days slice to reduce server load
        Default: 31

    Returns
    -------
    list
        A list of path2file
    """
    if not isinstance(website_name, str):
        raise TypeError("website_name should be a str type")

    if not isinstance(datamining_type, str):
        raise TypeError("datamining_type should be a str type")

    if not isinstance(payload, dict) or not payload:
        raise TypeError("payload should be a non-empty dict")

    if not isinstance(n_days_slice, int) or n_days_slice < 0:
        raise TypeError("n_days_slice should be a positive integer")

    l_allowed_datamining_types = ["order", "estimate", "isenginerequest", "actionlog", "scart"]

    if datamining_type not in l_allowed_datamining_types:
        raise ValueError(f"datamining_type={datamining_type} not allowed.\n\
                        Use one of the following: {', '.join(l_allowed_datamining_types)}")

    date_from = payload["date-from"] if "date-from" in payload else None
    if not date_from:
        raise ValueError("missing parameter=date-from in payload object")

    date_to = payload['date-to'] if 'date-to' in payload else None
    if not date_to:
        raise ValueError("missing parameter=date-from in payload object")

    date_format = "%m/%d/%Y"
    dt_date_from = datetime.strptime(date_from, date_format)
    dt_date_to = datetime.strptime(date_to, date_format)

    if dt_date_from > dt_date_to:
        raise ValueError("'date-from' cannot occur later than 'date-to'")

    # marketing attribution rule id, default to 0
    if "view-id" in payload:
        payload["view-id"] = str(payload["view-id"])
        match = re.match(
            pattern=r'^[0-9]$',
            string=payload["view-id"]
        )
        if not match:
            raise ValueError("view-id should match ^[0-9]$")

    else:
        payload["view-id"] = "0"

    website_name = website_name
    datamining_type = datamining_type
    payload = payload
    status_waiting_seconds = status_waiting_seconds
    n_days_slice = n_days_slice
    dt_tmp_date_to = dt_date_to
    n_days_slice = timedelta(days=n_days_slice)
    one_day_slice = timedelta(days=1)
    _os._create_directory(output_directory=output_directory)
    l_path2file = []  # store each file for n_days_slice
    # To avoid overloading the API with huge requests
    # We split the requests into smaller timeranres "n_days_slice"
    # While temporary date to delta is smaller than requested time to delta
    while dt_tmp_date_to <= dt_date_to:
        dt_tmp_date_to = dt_date_from + n_days_slice
        # Cannot request further than dt_date_to
        dt_tmp_date_to = dt_date_to if dt_tmp_date_to >= dt_date_to else dt_tmp_date_to
        date_from = dt_date_from.strftime(date_format)
        date_to = dt_tmp_date_to.strftime(date_format)
        date_from_file = date_from.replace("/", "_")
        date_to_file = date_to.replace("/", "_")

        output_filename = "_".join([
            website_name,
            datamining_type,
            "view",
            payload["view-id"],
            "from",
            date_from_file,
            "to",
            date_to_file,
        ]) + ".csv.gz"
        output_path2file = os.path.join(output_directory, output_filename)

        if not _request._is_skippable(
                output_path2file=output_path2file,
                override_file=override_file,
                print_log=self._print_log
        ):
            payload['date-from'] = date_from
            payload['date-to'] = date_to

            search_url = f"{self._api_v2}/ea/{website_name}/report/{datamining_type}/search.json"

            search_json = _request._to_json(
                request_type="get",
                url=search_url,
                params=payload,
                headers=self._http_headers,
                print_log=self._print_log
            )

            jobrun_id = search_json["jobrun_id"]

            status_url = f"{self._api_v2}/ea/{website_name}/report/{datamining_type}/status.json"
            status_payload = {"jobrun-id": jobrun_id}
            ready = False

            if not isinstance(status_waiting_seconds, int) or status_waiting_seconds < 5:
                status_waiting_seconds = 5

            while not ready:
                self._log(f'Waiting for jobrun_id={jobrun_id} to complete')
                time.sleep(status_waiting_seconds)
                status_json = _request._to_json(
                    request_type="get",
                    url=status_url,
                    params=status_payload,
                    headers=self._http_headers,
                    print_log=self._print_log
                )

                if status_json["jobrun_status"] == "COMPLETED":
                    ready = True

            download_url = f"{self._api_v2}/ea/{website_name}/report/{datamining_type}/download.json"
            download_payload = {'output-as-csv': 0, 'jobrun-id': jobrun_id}

            req = urllib.request.Request(
                url=f"{download_url}?{urllib.parse.urlencode(download_payload)}",
                headers=self._http_headers,
            )
            _stream_req(
                req=req,
                output_path2file=output_path2file)
        l_path2file.append(output_path2file)

        # last iteration, we queried up to the requested date
        if dt_tmp_date_to == dt_date_to:
            break

        # add one_day_slice to avoid querying the same day twice
        dt_date_from += n_days_slice + one_day_slice

    return l_path2file


def _stream_req(
    req: urllib.request.Request,
    output_path2file: str
) -> None:
    """ Stream datamining data in csv gzipped file

    Parameters
    ----------
    req: urllib.request.Request, obligatory
    output_path2file: str, obligatory
    """
    with gzip.open(
            filename=output_path2file,
            mode="wt"
    ) as csvfile:

        csvwriter = csv.writer(
            csvfile,
            delimiter=';'
        )

        with urllib.request.urlopen(req) as f:
            columns = []
            objects = ijson.items(f, "data.fields.item")  # .item is for ijson
            headers = (header for header in objects)
            # working on header.name rather than header.header for consitency
            # because the latest is language specific
            for header in headers:
                # this allows to recover the name of the product param name
                # instead of the id
                if header["name"].startswith('productparam_'):
                    match = re.search(
                        pattern=r'\s:\s([\w\W]+?)\s#\s(\d+)$',
                        string=header["header"]
                    )
                    prdp_name = match.group(1)
                    prd_idx = int(match.group(2)) - 1  # start at 0
                    header["name"] = f"productparam_{prdp_name}_{prd_idx}"
                # this allows to recover the name of the cgi param name
                # instead of the id
                if header["name"].startswith('cgiparam_'):
                    match = re.search(
                        pattern=r'^[\w\W]+?\s:\s(.*)$',
                        string=header["header"],
                    )
                    cgip_name = match.group(1)
                    header["name"] = f"cgiparam_{cgip_name}"
                # this allows to recover the name of the CRM param name
                # instead of the id
                if header["name"].startswith('iduserparam_'):
                    match = re.search(
                        pattern=r'^[\w\W]+?\s:\s(.*)$',
                        string=header["header"],
                    )
                    iduserparam_name = match.group(1)
                    header["name"] = f"iduserparam_{iduserparam_name}"
                # this allows to recover the name of the audience name
                # instead of the id
                if header["name"].startswith('cluster_'):
                    match = re.search(
                        pattern=r'^[\w\W]+?\s:\s(.*)$',
                        string=header["header"],
                    )
                    cluster_name = match.group(1)
                    header["name"] = f"cluster_{cluster_name}"

                columns.append(header["name"])
            csvwriter.writerow(columns)

        with urllib.request.urlopen(req) as f:
            objects = ijson.items(f, "data.rows.item")  # .item is for ijson
            rows = (row for row in objects)
            for row in rows:
                csvwriter.writerow(row)



"""This module allows to manipulate a pandas DataFrame"""

import pandas as _pd

def move_columns(
    df : _pd.DataFrame,
    colname_idx : dict,
    inplace : bool = False,
) -> _pd.DataFrame:
    """ For a given DataFrame, move given column_name keys into the given index values

    Parameters
    ----------
    df : pd.DataFrame, mandatory

    colname_idx: dict, mandatory
        { "colname1" : 2, "colname10" : 1 }

    Returns
    -------
    pf.DataFrame with updated columns
    """

    if not inplace:
        df=df.copy()

    if not isinstance(df, _pd.DataFrame):
        raise TypeError("df should be a pandas DataFrame")

    if not isinstance(colname_idx, dict):
        raise TypeError("colname_idx should be a dict")

    for column_name in colname_idx.keys():
        if not isinstance(column_name, str):
            raise TypeError(f"key={column_name}, should be an str type")

        if column_name not in df.columns:
            raise ValueError(f"key={column_name} is not a column of the given DataFrame")

        idx = colname_idx[column_name]

        if not isinstance(idx, int):
            raise TypeError(f"key={column_name}, val={idx} should be an integer type")

        if df.columns.get_loc(column_name) == idx:
            next

        tmp_col = df[column_name]

        df.drop(
            labels=[column_name],
            axis=1,
            inplace=True
        )

        df.insert(
            loc=idx,
            column=column_name,
            value=tmp_col
        )

        return df

def query_string_to_column(
    df : _pd.DataFrame,
    col_name : list,
    n_decode : int =  0,
) -> _pd.DataFrame:
    """ For a given DataFrame, move given column_name keys into the given index values

    Parameters
    ----------
    df : pd.DataFrame, mandatory

    col_name: list, mandatory
       The columns in which to retrieve the query strings

    Returns
    -------
    pf.DataFrame with updated columns
    """


"""This module allows to download flat realtime report data
from the Eulerian Technologies API"""

import pandas as pd

from eanalytics_api_py.internal import _request


def download_flat_overview_realtime_report(
        self,
        date_from: str,
        date_to: str,
        website_name: str,
        report_name: list,
        kpi: list,
        channel: list = None,
        view_id: int = 0,
        mdevicetype_id: list = None,
        ordertype_id: list = None,
        ordertypecustom_id: list = None,
        profile_id: list = None,
) -> pd.DataFrame:
    """ Fetch realtime report data into a pandas dataframe

    Parameters
    ----------
    date_from: str, mandatory
        mm/dd/yyyy

    date_to: str, mandatory
        mm/dd/yyyy

    website_name: str, mandatory
        Your targeted website_name in Eulerian Technologies platform

    report_name: str, mandatory

    kpi: list, mandatory
        List of kpis to request

    channel: list, mandatory
        List of channels (ADVERTISING...)

    view_id: int, optional
        Between 0 and 9

    mdevicetype_id: list, optional
        List of mdevicetype_id to filter on

    ordertype_id: list, optional
        List of ordertype_id to filter on

    ordertypecustom_id: list, optional
        List of ordertypecustom_id to filter on

    profile_id: list, optional
        List of profile_id to filter on

    Returns
    -------
    pd.DataFrame()
        A pandas dataframe
    """

    if not isinstance(date_from, str):
        raise TypeError("date_from should be a string dtype")

    if not isinstance(date_to, str):
        raise TypeError("date_to should be a string dtype")

    if not isinstance(website_name, str):
        raise TypeError("website_name should be a string dtype")

    if not isinstance(report_name, str):
        raise TypeError("report_name should be a string dtype")

    if not isinstance(kpi, list):
        raise TypeError("kpi should be a list dtype")

    payload = {
        'date-from': date_from,
        'date-to': date_to,
    }
    if mdevicetype_id:
        if not isinstance(mdevicetype_id, list):
            raise TypeError("mdevicetype_id should be a list")

        mdevicetype_map = self.get_mdevicetype_id_name_map(website_name)
        for _id in mdevicetype_id:
            if _id not in mdevicetype_map:
                raise ValueError(f"mdevicetype_id={_id} not allowed. Allowed={', '.join(mdevicetype_map.keys())}")
        payload["ea-subk"] = ",".join(mdevicetype_id)

    if ordertype_id:
        if not isinstance(ordertype_id, list):
            raise TypeError("ordertype_id should be a list")
        ordertype_map = self.get_ordertype_id_name_map(website_name)
        for _id in ordertype_id:
            if _id not in ordertype_map:
                raise ValueError(f"ordertype_id={_id} not allowed. Allowed={', '.join(ordertype_map.keys())}")
        payload["shoppingcart-k1"] = ",".join(ordertype_id)

    if ordertypecustom_id:
        if not isinstance(ordertypecustom_id, list):
            raise TypeError("ordertypecustom_id should be a list")
        ordertypecustom_map = self.get_ordertypecustom_id_name_map(website_name)
        for _id in ordertypecustom_id:
            if _id not in ordertypecustom_map:
                raise ValueError(
                    f"ordertypecustom_id={_id} not allowed. Allowed={', '.join(ordertypecustom_map.keys())}")
        payload["shoppingcart-k1-custom"] = ",".join(ordertypecustom_id)

    if profile_id:
        if not isinstance(profile_id, list):
            raise TypeError("profile_id should be a list")
        profile_map = self.get_profile_id_name_map(website_name)
        for _id in profile_id:
            if _id not in profile_map:
                raise ValueError(f"profile_id={_id} not allowed. Allowed={', '.join(profile_map.keys())}")
        payload["visitpprofil-k1"] = ",".join(profile_id)

    view_id = str(view_id)
    view_map = self.get_view_id_name_map(website_name)
    if view_id not in view_map:
        raise ValueError(f"view_id={view_id} not found. Allowed: {', '.join(view_map.keys())}")

    payload["view-id"] = view_id

    d_website = self.get_website_by_name(website_name)
    url = f"{self._api_v2}/ea/{website_name}/report/realtime/{report_name}.json"
    path_module = __import__(
        name="eanalytics_api_py.internal.realtime_overview.path._" + report_name,
        fromlist=report_name)

    if not channel:
        channel = list(path_module.d_path.keys())

    l_df = []
    d_path = path_module.d_path.copy()  # because we override values we want a clean copy
    for _channel in channel:
        l_path = d_path[_channel]["path"]
        l_path[0] = l_path[0] % int(d_website["website_id"])

        l_dim = d_path[_channel]["dim"]
        payload['path'] = ".".join(l_path)
        payload['ea-columns'] = ",".join([*l_dim, *kpi])

        _json = _request._to_json(
            url=url,
            request_type="get",
            params=payload,
            headers=self._http_headers,
            print_log=True
        )
        sub_df = pd.DataFrame(
            data=_json["data"]["rows"],
            columns=[d_field["name"] for d_field in _json["data"]["fields"]])

        l_df.append(sub_df)

    df = pd.concat(
        l_df,
        axis=0,
        ignore_index=True)

    for name, alias in path_module.override_dim_map.items():
        mask = (df[alias].isin([0, '0']))
        df.loc[mask, alias] = df[name]
        df.drop(
            labels=alias,
            axis=1,
            inplace=True)

    df.rename(
        columns=path_module.dim_px_map,
        inplace=True)

    return df
